@@ -2,7 +2,6 @@ import logging
 import sys
 import threading
 from ctypes import byref, c_int, c_ubyte, pointer
-from pathlib import Path
 from queue import Empty, Queue
 from time import sleep, time
 from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
@@ -11,8 +10,6 @@ from can import BusABC, CanProtocol, Message
 from can.broadcastmanager import CyclicSendTaskABC
 from can.typechecking import AutoDetectedConfig, CanFilters
 
-# TODO: REMOVE
-sys.path.append(Path(__file__).resolve().parents[1].as_posix())
 from can_cando.CANdoImport import (
     CAN_MSG_TIMESTAMP_RESOLUTION,
     CANDO_CAN_BUFFER_LENGTH,
@@ -77,6 +74,56 @@ else:
 
 
 class CANdoBus(BusABC):
+    """Docs from CANdo SDK Programmer's Guide.
+
+    The CANdoGetDevices(...) function returns a list of all the CANdo devices connected to the PC,
+    identifying each by hardware type & serial number.
+
+    Either the CANdoOpen(…) or CANdoOpenDevice(...) function must be called before using a CANdo device, in order to
+    open a connection to the device. These functions attempt to find a free CANdo on the PC, returning a handle
+    to the device if successful. This handle is part of the TCANdoUSB structure passed into all the other API functions.
+    NOTE : A valid handle must be obtained by calling one of these functions before calling any of the other SDK
+    functions, except for the functions CANdoGetPID(…), CANdoGetDevices(...) & CANdoGetVersion(...), as these don't
+    require a handle. The CANdoOpen(…) function returns a connection to the first free CANdo device found, regardless
+    of hardware type or serial number. The CANdoOpenDevice(…) function allows a particular CANdo device to be selected
+    based on hardware type, serial number or both. A single entry in the CANdoDevices array returned by the
+    CANdoGetDevices(…) function maybe passed into the CANdoOpenDevice(…) function to select a particular device.
+
+    After obtaining a handle, the CANdo device is initialised by calls to
+    the CANdoSetBaudRate(…), CANdoSetMode(…) & CANdoSetFilters(…) functions.
+
+    Following initialisation, the CANdoSetState(…) function is called to put CANdo into run mode, enabling
+    reception & transmission of CAN messages & starting the internal CAN message timestamp.
+
+    The CANdoReceive(…) & CANdoTransmit(…) functions, receive & transmit CAN messages respectively.
+
+    The CANdoRequestStatus(…) function requests the internal status of the CANdo device & the CAN bus. If any errors
+    are detected in the operation of the CANdo device or an error occurs on the CAN bus, then one or more status
+    messages are automatically generated by the CANdo device & returned to the PC. The status is returned within the
+    TCANdoStatus structure (see Appendix A), which is populated by a call to the CANdoReceive(...) function.
+
+    The CANdoRequestDateStatus(…) function requests the date of manufacture of the device. The status is returned in
+    the same manner as for the CANdoRequestStatus(...) function.
+
+    The CANdoRequestBusLoadStatus(…) function requests the CAN bus message load on the connected CAN bus. The status is
+    returned in the same manner as for the CANdoRequestStatus(...) function.
+
+    The CANdoRequestSetupStatus(...) function requests the CAN baud rate & CAN operating mode of the device (CANdo AUTO
+    device only). The status is returned in the same manner as for the CANdoRequestStatus(...) function.
+
+    The CANdoClearStatus(…) function clears any pending system status message within the CANdo device. CAN bus related
+    errors are not cleared by this command, as these are handled by the low level CAN controller.
+
+    The CANdoGetVersion(…) function returns the versions of the API library,
+    the USB library & the USB driver installed on the PC.
+
+    The reception & transmission of CAN messages is stopped by a further
+    call to the CANdoSetState(…) function with the state set to stop.
+
+    The CANdoClose(…) function must be called before terminating an application,
+    to close the CANdo connection & free the handle to the device.
+    """
+
     # No. of devices to look for
     NoOfDevices: c_int
 
@@ -396,6 +443,21 @@ class CANdoBus(BusABC):
         :param Rx2IDE4: - flag to indicate Rx2Filter4 value is either an 11 or 29 bit value (0 = 11 bit ID, 1 = 29 bit ID)
         :param Rx2Filter4: - receive buffer 2, filter 4 (See Appendix B for further details)
         """
+        log.debug("Setting CANdo(ISO) filters:")
+        log.debug(f"Rx1Mask =    {hex(Rx1Mask)}")
+        log.debug(f"Rx1IDE1 =    {hex(Rx1IDE1)}")
+        log.debug(f"Rx1Filter1 = {hex(Rx1Filter1)}")
+        log.debug(f"Rx1IDE2 =    {hex(Rx1IDE2)}")
+        log.debug(f"Rx1Filter2 = {hex(Rx1Filter2)}")
+        log.debug(f"Rx2Mask =    {hex(Rx2Mask)}")
+        log.debug(f"Rx2IDE1 =    {hex(Rx2IDE1)}")
+        log.debug(f"Rx2Filter1 = {hex(Rx2Filter1)}")
+        log.debug(f"Rx2IDE2 =    {hex(Rx2IDE2)}")
+        log.debug(f"Rx2Filter2 = {hex(Rx2Filter2)}")
+        log.debug(f"Rx2IDE3 =    {hex(Rx2IDE3)}")
+        log.debug(f"Rx2Filter3 = {hex(Rx2Filter3)}")
+        log.debug(f"Rx2IDE4 =    {hex(Rx2IDE4)}")
+        log.debug(f"Rx2Filter4 = {hex(Rx2Filter4)}")
         ret = CANdoSetFilters(
             self.CANdoUSBPtr,
             Rx1Mask,
@@ -653,58 +715,68 @@ class CANdoBus(BusABC):
                 log.error(f"Failed to stop transmission on buffer {buf_no}, error code: {CANdoFuncRetCodeErr(ret)}")
 
     def _recv_t(self) -> None:
-        # Local copy to avoid global lookups
-        ts_resolution = CAN_MSG_TIMESTAMP_RESOLUTION
+        # Docstring is for "CANdoReceive" function
+        """Receive CAN & status messages sent by CANdo.
 
+        The CAN receive message cyclic buffer is passed into the function via the CANdoCANBufferPointer parameter.
+        The CANdoReceive(…) function populates this cyclic buffer with any CAN messages received via the CAN bus.
+        A single status message is returned by the CANdoReceive(…) function, if an error occurs or a specific status
+        request is sent using one of the functions, CANdoRequestStatus(…), CANdoRequestDateStatus(…), CANdoRequestBusLoadStatus(…),
+        CANdoRequestSetupStatus(...) or CANdoRequestAnalogInputStatus(…). The content of the status message
+        depends upon the type of status message returned by the device & is described in detail
+        in Appendix A, table 'TCANdoStatus Content Versus NewFlag Value'.
+        """
         while not self._is_shutdown:
-            # Clear status flag
-            self.CANdoStatus.NewFlag = CANDO_NO_STATUS
+            has_received_messages = False
 
             with self._recv_lock:  # Do not interfere with status requests
-                if CANdoReceive(self.CANdoUSBPtr, self.CANdoCANBufferPtr, self.CANdoStatusPtr) != CANdoFuncRetCode.CANDO_SUCCESS:
-                    raise CANdoOperationError("Error receiving messages")
+                self.CANdoStatus.NewFlag = CANDO_NO_STATUS  # Clear status flag
 
-                # print(
-                #     f"Status: {CANdoStsErr(self.CANdoStatus.Status)}, "
-                #     f"Bus State: {CANdoBusStsErr(self.CANdoStatus.BusState)}",
-                # )
+                ret = CANdoReceive(self.CANdoUSBPtr, self.CANdoCANBufferPtr, self.CANdoStatusPtr)
+                if ret != CANdoFuncRetCode.CANDO_SUCCESS:
+                    raise CANdoOperationError(f"Failed to receive messages, error code: {CANdoFuncRetCodeErr(ret)}")
 
-            if self.CANdoCANBuffer.ReadIndex == self.CANdoCANBuffer.WriteIndex and not self.CANdoCANBuffer.FullFlag:
-                # No messages were received, sleep for a short time
+                # log.debug(f"Status: {CANdoStsErr(self.CANdoStatus.Status)}, \nBus State: {CANdoBusStsErr(self.CANdoStatus.BusState)}")
+
+                while self.CANdoCANBuffer.ReadIndex != self.CANdoCANBuffer.WriteIndex or self.CANdoCANBuffer.FullFlag:
+                    if not has_received_messages:
+                        has_received_messages = True
+                        if self.CANdoCANBuffer.ReadIndex < self.CANdoCANBuffer.WriteIndex:
+                            n_messages = self.CANdoCANBuffer.WriteIndex - self.CANdoCANBuffer.ReadIndex
+                        else:
+                            n_messages = CANDO_CAN_BUFFER_LENGTH - self.CANdoCANBuffer.ReadIndex
+                        log.debug(f'Received {n_messages:4} messages in the last "CANdoReceive(...)" call.')
+
+                    arbitration_id = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].ID
+                    ide = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].IDE
+                    rtr = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].RTR
+                    dlc = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].DLC
+                    data = bytearray(self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].Data[:dlc])
+
+                    # # Unused for now; however, check for the "state" property and "BusState" enum  in bus.py
+                    # bus_state = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].BusState
+
+                    # Timestamp - a timestamp indicating the receive time of the message since starting CANdo.
+                    timestamp = (
+                        self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].TimeStamp * CAN_MSG_TIMESTAMP_RESOLUTION
+                    ) + self.t_start
+
+                    # Increment index onto next circular buffer element
+                    self.CANdoCANBuffer.ReadIndex = (self.CANdoCANBuffer.ReadIndex + 1) % CANDO_CAN_BUFFER_LENGTH
+                    self.CANdoCANBuffer.FullFlag = False  # Buffer is no longer full
+
+                    self.msg_queue.put((timestamp, arbitration_id, bool(ide), rtr, dlc, data))
+
+                    if self.CANdoStatus.NewFlag == CANDO_DEVICE_STATUS:  # type: ignore
+                        # log.debug(f"  Status = {self.CANdoStatus.Status}, Bus State = {self.CANdoStatus.BusState}")
+                        self.CANdoStatus.NewFlag = CANDO_NO_STATUS  # Clear status flag if status message received
+
+            if has_received_messages:
+                if self.CANdoStatus.Status != CANDO_NO_STATUS:
+                    log.debug(f"Status: {CANdoStsErr(self.CANdoStatus.Status)}")
+            else:
+                # No messages were received, sleep for a short time OUTSIDE the lock
                 sleep(0.01)
-                continue
-
-            while self.CANdoCANBuffer.ReadIndex != self.CANdoCANBuffer.WriteIndex or self.CANdoCANBuffer.FullFlag:
-                arbitration_id = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].ID
-                ide = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].IDE
-                rtr = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].RTR
-                dlc = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].DLC
-                data = bytearray(self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].Data[:dlc])
-
-                # # Unused for now; however, check for the "state" property and "BusState" enum  in bus.py
-                # bus_state = self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].BusState
-
-                # Timestamp - a timestamp indicating the receive time of the message since starting CANdo.
-                timestamp = (self.CANdoCANBuffer.CANMessage[self.CANdoCANBuffer.ReadIndex].TimeStamp * ts_resolution) + self.t_start
-
-                # Increment index onto next circular buffer element
-                self.CANdoCANBuffer.ReadIndex = (self.CANdoCANBuffer.ReadIndex + 1) % CANDO_CAN_BUFFER_LENGTH
-                self.CANdoCANBuffer.FullFlag = False  # No longer full
-
-                self.msg_queue.put(
-                    (
-                        timestamp,
-                        arbitration_id,
-                        bool(ide),
-                        rtr,
-                        dlc,
-                        data,
-                    ),
-                )
-
-                if self.CANdoStatus.NewFlag == CANDO_DEVICE_STATUS:  # type: ignore
-                    # log.debug(f"  Status = {self.CANdoStatus.Status}, Bus State = {self.CANdoStatus.BusState}")
-                    self.CANdoStatus.NewFlag = CANDO_NO_STATUS  # Clear status flag
 
     @override
     def flush_tx_buffer(self) -> None:
@@ -720,7 +792,7 @@ class CANdoBus(BusABC):
         if bus.NoOfDevices.value > 0:
             interfaces = [AutoDetectedConfig(interface="cando", channel=c) for c in range(bus.NoOfDevices.value)]
 
-        bus.shutdown()  # Close connection to the mo
+        bus.shutdown()  # Close connection
         del bus
 
         return interfaces
@@ -728,16 +800,13 @@ class CANdoBus(BusABC):
     @override
     def send(self, msg: Message, timeout: Optional[float] = None) -> None:
         if not self.CANdoUSB.OpenFlag:
-            raise CANdoOperationError("Device not open!")
-
-        # DataArray = c_ubyte * msg.dlc  # Create a data array type
+            raise CANdoOperationError("CANdo(ISO) device is not open!")
 
         tx_res = self.cando_transmit(
             int(msg.is_extended_id),
             msg.arbitration_id,
             int(msg.is_remote_frame),
             msg.dlc,
-            # DataArray(*msg.data),  # Create & initialise an instance of the data array
             self.DataArrays[msg.dlc](*msg.data),  # Create & initialise an instance of the data array
         )
 
@@ -856,7 +925,11 @@ class CANdoBus(BusABC):
 
     @override
     def shutdown(self) -> None:
-        super().shutdown()
+        try:
+            super().shutdown()
+        except Exception:
+            log.exception("Error during shutdown!")
+
         if hasattr(self, "_recv_thread") and self._recv_thread.is_alive():
             self._recv_thread.join()
 
@@ -889,51 +962,6 @@ class CANdoBus(BusABC):
 
     @override
     def stop_all_periodic_tasks(self, remove_tasks: bool = True) -> None:
-        self.cando_stop_all_periodic_transmissions()
+        if hasattr(self, "CANdoUSB"):
+            self.cando_stop_all_periodic_transmissions()
         super().stop_all_periodic_tasks(remove_tasks=remove_tasks)
-
-
-def main() -> None:
-    interfaces = CANdoBus._detect_available_configs()  # type: ignore
-
-    if len(interfaces) == 0:
-        print("No CANdo(ISO) devices detected!")
-        return
-
-    print("Available CANdo(ISO) channels:")
-    for i in interfaces:
-        print(f"  {i}")
-    selected = int(input("Select channel: ") or "0")
-
-    bus = CANdoBus(selected)
-
-    try:
-        while True:
-            _in = input(
-                "Press 't' to transmit a message, 'r' to receive a message,'s' to see the device's status and 'q' or 'Ctrl-C' to quit: ",
-            )
-            if _in == "q":
-                break
-            if _in == "s":
-                bus.print_status()
-            if _in == "t":
-                bus.send(
-                    Message(
-                        arbitration_id=0x1500,
-                        data=[1, 2, 3, 4, 5, 6, 7, 8],
-                        is_extended_id=True,
-                    ),
-                )
-            if _in == "r":
-                msg = bus.recv(timeout=1)
-                if msg is not None:
-                    print(msg)
-                else:
-                    print("No message received...")
-    except KeyboardInterrupt:
-        pass
-    bus.shutdown()
-
-
-if __name__ == "__main__":
-    main()
